"""High level Object-Oriented interface methods for the package."""

_author__ = "Daniel da Silva <mail@danieldasilva.org>"

import numpy as np
from .decode import _decode_fixed_length


class PacketField(object):
    """A field contained in a packet.
    """
    
    def __init__(self, name, data_type, bit_length, bit_offset=None,
                 byte_order='big'):
        """Define a new PacketField.
        
        Parameters
        ----------
        name : str
            String identifier for the field. The name specified how you may
            call upon this data later.
        data_type : {'uint', 'int', 'float', 'str', 'fill'}
            Data type of the field.
        bit_length : int
            Number of bits contained in the field.
        bit_offset : int, optional
            Bit offset into packet, including primary header. If this is not
            specified, than the bit offset will the be calculated automatically
            from its position inside the packet definition.
        byte_order : {'big', 'little'}, optional
            Byte order of the field. Defaults to big endian.

        Raises
        ------
        TypeError
             If one of the arguments is not of the correct type.
        ValueError
             data_type or byte_order is invalid
        """
        if not isinstance(name, str):
            raise TypeError('name parameter must be a str')
        if not isinstance(data_type, str):
            raise TypeError('data_type parameter must be a str')
        if not isinstance(bit_length, (int, np.integer)):
            raise TypeError('bit_length parameter must be an int')
        if not (bit_offset is None or isinstance(bit_offset, (int, np.integer))):
            raise TypeError('bit_offset parameter must be an int')
        
        valid_data_types = ('uint', 'int', 'float', 'str', 'fill')
        if data_type not in valid_data_types:
            raise ValueError('data_type must be one of {valids}'.format(
                valids=repr(valid_data_types)))

        valid_byte_orders = ('big', 'little')
        if byte_order not in valid_byte_orders:
            raise ValueError('byte_order must be one of {valids}'.format(
                valids=repr(valid_byte_orders)))        
        
        self._name = name
        self._data_type = data_type
        self._bit_length = bit_length
        self._bit_offset = bit_offset
        self._byte_order = byte_order

    def __repr__(self):
        values = {k: repr(v) for (k, v) in self.__dict__.items()}

        return ('PacketField(name={_name}, data_type={_data_type}, '
                'bit_length={_bit_length}, bit_offset={_bit_offset}, '
                'byte_order={_byte_order})'.format(**values))

                
class FixedLength(object):
    """Define a fixed length packet to decode binary data.

    In the context of engineering and science, fixed length packets correspond
    to data that is of the same layout every time. Examples of this include
    sensor time series, status codes, or error messages.
    """    
    def __init__(self, fields):
        """        
        Parameters
        ----------
        fields : list of `ccsdspy.PacketField`
            Layout of packet fields contained in the definition.
        """
        self._fields = fields[:]
       
    def load(self, file):
        """Decode a file-like object containing a sequence of these packets.

        Parameters
        ----------
        file: str
           Path to file on the local file system, or file-like object

        Returns
        -------
        `OrderedDict` mapping field names to NumPy arrays.
        """
        if hasattr(file, 'read'):
            file_bytes = np.fromstring(file.read(), 'u1')
        else:
            file_bytes = np.fromfile(file, 'u1')

        field_arrays = _decode_fixed_length(file_bytes, self._fields)
        return field_arrays




# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
IO Interface for Reading CCSDS Data in Python.
"""

# For egg_info test builds to pass, put package imports here.

from .interface import (FixedLength, PacketField)







"""Internal decoding routines."""
from __future__ import division
from collections import namedtuple, OrderedDict
import numpy as np

__author__ = 'Daniel da Silva <mail@danieldasilva.org>'


def _decode_fixed_length(file_bytes, fields):
    """Decode a fixed length APID.
    
    Parameters
    ----------
    file_bytes : array 
       A NumPy array of uint8 type, holding the bytes of the file to decode.
    fields : list of ccsdspy.interface.PacketField
       A list of fields, including the secondary header but excluding the
       primary header.

    Returns
    -------
    Ordered dictionary mapping field names to NumPy arrays.
    """
    # Setup a dictionary mapping a bit offset to each field. It is assumed
    # that the `fields` array contains entries for the secondary header.
    packet_nbytes = file_bytes[4] * 256 + file_bytes[5] + 7
    body_nbytes = sum(field._bit_length for field in fields) // 8
    counter = (packet_nbytes - body_nbytes) * 8
    
    bit_offset = {}

    for field in fields:
        if field._bit_offset is None:
            bit_offset[field._name] = counter
            counter += field._bit_length
        else:
            bit_offset[field._name] = field._bit_offset

    if all(field._bit_offset is None for field in fields):
        assert counter == packet_nbytes * 8, \
            'Field definition != packet length'.format(n=counter-packet_nbytes*8)
        
    # Setup metadata for each field, consiting of where to look for the field in
    # the file and how to parse it.
    FieldMeta = namedtuple('Meta', ['nbytes_file', 'start_byte_file',
                                    'nbytes_final', 'np_dtype'])
    field_meta = {}

    for field in fields:
        nbytes_file = np.ceil(field._bit_length/8.).astype(int)

        if (bit_offset[field._name] % 8 and
             bit_offset[field._name] % 8 + field._bit_length > 8):
            nbytes_file += 1

        nbytes_final = {3: 4, 5: 8, 6: 8, 7: 8}.get(nbytes_file,  nbytes_file)
        start_byte_file = bit_offset[field._name] // 8

        np_dtype = {
            'uint': '>u%d' % nbytes_final,
            'int':  '>i%d' % nbytes_final,
            'fill': '>u%d' % nbytes_final,
            'float': 'f%d' % nbytes_final,
            'str':   'S%d' % nbytes_final,
        }[field._data_type]
        
        field_meta[field] = FieldMeta(
            nbytes_file, start_byte_file, nbytes_final, np_dtype)

    # Read the file and calculate length of packet and number of packets in the
    # file. Trim extra bytes that may have occurred by a break in the downlink
    # while a packet was beign transferred.
    extra_bytes = file_bytes.size % packet_nbytes

    if extra_bytes > 0:
        file_bytes = file_bytes[:-extra_bytes]
    
    packet_count = file_bytes.size // packet_nbytes
    
    # Create byte arrays for each field. At the end of this method they are left
    # as the numpy uint8 type.
    field_bytes = {}

    for field in fields:
        meta = field_meta[field]
        arr = np.zeros(packet_count * meta.nbytes_final, 'u1')
        xbytes = meta.nbytes_final - meta.nbytes_file

        for i in range(xbytes, meta.nbytes_final):
            arr[i::meta.nbytes_final] = (
                file_bytes[meta.start_byte_file + i - xbytes::packet_nbytes]
            )
            field_bytes[field] = arr

    # Switch dtype of byte arrays to the final dtype, and apply masks and shifts
    # to interpret the correct bits.
    field_arrays = OrderedDict()

    for field in fields:
        meta = field_meta[field]
        arr = field_bytes[field]
        arr.dtype = meta.np_dtype

        if field._data_type in ('int', 'uint'):
            xbytes = meta.nbytes_final - meta.nbytes_file

            bitmask_left = (bit_offset[field._name]
                            + 8 * xbytes
                            - 8 * meta.start_byte_file)

            bitmask_right = (8 * meta.nbytes_final
                             - bitmask_left
                             - field._bit_length)
     
            bitmask_left, bitmask_right = (
                np.array([bitmask_left, bitmask_right]).astype(meta.np_dtype)
            )
            
            bitmask = np.zeros(arr.shape, meta.np_dtype)
            bitmask |= (1 << int(8 * meta.nbytes_final - bitmask_left)) - 1
            tmp = np.left_shift([1], bitmask_right)
            bitmask &= np.bitwise_not(tmp[0] - 1).astype(meta.np_dtype)
        
            arr &= bitmask
            arr >>= bitmask_right
            
            if field._byte_order == 'little':
                arr.byteswap(inplace=True)

        field_arrays[field._name] = arr

    return field_arrays



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst
from setuptools import setup

with open("README.rst") as f:
    LONG_DESCRIPTION = f.read()

PACKAGENAME = 'ccsdspy'
DESCRIPTION = 'IO Interface for Reading CCSDS Data in Python.'
AUTHOR = 'Daniel da Silva'
AUTHOR_EMAIL = ''
LICENSE = 'unknown'
URL = 'http://github.com/ddasilva/ccsdspy'
# VERSION should be PEP440 compatible (http://www.python.org/dev/peps/pep-0440)
VERSION = '0.0.8'

setup(
    name=PACKAGENAME,
    version=VERSION,
    description=DESCRIPTION,
    install_requires=[
        "numpy"
    ],
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    url=URL,
    packages=['ccsdspy'],
    long_description=LONG_DESCRIPTION,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable  ",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ]
)
